@@ -1,6 +1,28 @@
 #include "confuse.h"
 #include <string.h>
 
+void print_func(cfg_opt_t *opt, unsigned int index, FILE *fp)
+{
+	fprintf(fp, "%s(foo)", opt->name);
+}
+
+void print_ask(cfg_opt_t *opt, unsigned int index, FILE *fp)
+{
+	int value = cfg_opt_getnint(opt, index);
+	switch(value) {
+		case 1:
+			fprintf(fp, "yes");
+			break;
+		case 2:
+			fprintf(fp, "no");
+			break;
+		case 3:
+		default:
+			fprintf(fp, "maybe");
+			break;
+	}
+}
+
 /* function callback
  */
 int cb_func(cfg_t *cfg, cfg_opt_t *opt, int argc, const char **argv)
@@ -40,7 +62,25 @@ int cb_verify_ask(cfg_t *cfg, cfg_opt_t *opt, const char *value, void *result)
 	return 0;
 }
 
-int main(void)
+int cb_validate_bookmark(cfg_t *cfg, cfg_opt_t *opt)
+{
+	/* only validate the last bookmark */
+	cfg_t *sec = cfg_opt_getnsec(opt, cfg_opt_size(opt) - 1);
+	if(!sec)
+	{
+		cfg_error(cfg, "section is NULL!?");
+		return -1;
+	}
+	if(cfg_getstr(sec, "machine") == 0)
+	{
+		cfg_error(cfg, "machine option must be set for bookmark '%s'",
+				  cfg_title(sec));
+		return -1;
+	}
+	return 0;
+}
+
+int main(int argc, char **argv)
 {
 	unsigned int i;
 	cfg_t *cfg;
@@ -50,7 +90,7 @@ int main(void)
 		CFG_INT("type", 0, CFGF_NONE),
 		CFG_STR("host", 0, CFGF_NONE),
 		CFG_STR_LIST("exclude", "{localhost, .localnet}", CFGF_NONE),
-		CFG_INT("port", 0, CFGF_NONE),
+		CFG_INT("port", 21, CFGF_NONE),
 		CFG_END()
 	};
 	static cfg_opt_t bookmark_opts[] = {
@@ -68,11 +108,11 @@ int main(void)
 		CFG_STR("probe-device", "eth2", CFGF_NONE),
 		CFG_SEC("bookmark", bookmark_opts, CFGF_MULTI | CFGF_TITLE),
 		CFG_FLOAT_LIST("delays", "{3.567e2, 0.2, -47.11}", CFGF_NONE),
-		{"func",CFGT_FUNC,0,0,CFGF_NONE,0,{0,0,cfg_false,0,"func(default, value) func(second,default)"},&cb_func,0,0},
-/*		CFG_FUNC("func", &cb_func),*/
+		CFG_FUNC("func", &cb_func),
 		CFG_INT_CB("ask-quit", 3, CFGF_NONE, &cb_verify_ask),
 		CFG_INT_LIST_CB("ask-quit-array", "{maybe, yes, no}",
 						CFGF_NONE, &cb_verify_ask),
+		CFG_FUNC("include", &cfg_include),
 		CFG_END()
 	};
 
@@ -82,7 +122,11 @@ int main(void)
 #endif
 
 	cfg = cfg_init(opts, CFGF_NOCASE);
-	ret = cfg_parse(cfg, "test.conf");
+
+	/* set a validating callback function for bookmark sections */
+	cfg_set_validate_func(cfg, "bookmark", &cb_validate_bookmark);
+
+	ret = cfg_parse(cfg, argc > 1 ? argv[1] : "test.conf");
 	printf("ret == %d\n", ret);
 	if(ret == CFG_FILE_ERROR) {
 		perror("test.conf");
@@ -101,6 +145,7 @@ int main(void)
 	n = cfg_size(cfg, "bookmark");
 	printf("%d configured bookmarks:\n", n);
 	for(i = 0; i < n; i++) {
+		cfg_t *pxy;
 		cfg_t *bm = cfg_getnsec(cfg, "bookmark", i);
 		printf("  bookmark #%u (%s):\n", i+1, cfg_title(bm));
 		printf("    machine = %s\n", cfg_getstr(bm, "machine"));
@@ -111,12 +156,16 @@ int main(void)
 		printf("    directory = %s\n", cfg_getstr(bm, "directory"));
 		printf("    password = %s\n", cfg_getstr(bm, "password"));
 
-		if(cfg_size(bm, "proxy")) {
+		pxy = cfg_getsec(bm, "proxy");
+		if(pxy) {
 			int j, m;
-			cfg_t *pxy = cfg_getsec(bm, "proxy");
 			if(cfg_getstr(pxy, "host") == 0) {
-				printf("no proxy host is set, setting it to 'localhost'...\n");
-				cfg_setstr(pxy, "host", "localhost");
+				printf("      no proxy host is set, setting it to 'localhost'...\n");
+				/* For sections without CFGF_MULTI flag set, there is
+				 * also an extended syntax to get an option in a
+				 * subsection:
+				 */
+				cfg_setstr(bm, "proxy|host", "localhost");
 			}
 			printf("      proxy host is %s\n", cfg_getstr(pxy, "host"));
 			printf("      proxy type is %ld\n", cfg_getint(pxy, "type"));
@@ -139,17 +188,35 @@ int main(void)
 
 	/* Using cfg_setint(), the integer value for the option ask-quit
 	 * is not verified by the value parsing callback.
-	 */
-	cfg_setint(cfg, "ask-quit", 4);
+	 *
+	 *
+	 cfg_setint(cfg, "ask-quit", 4);
+	 printf("ask-quit == %ld\n", cfg_getint(cfg, "ask-quit"));
+	*/
 
-	printf("ask-quit == %ld\n", cfg_getint(cfg, "ask-quit"));
+	/* The following commented line will generate a failed assertion
+	 * and abort, since the option "foo" is not declared
+	 *
+	 *
+	 printf("foo == %ld\n", cfg_getint(cfg, "foo"));
+	*/
 
-	cfg_addlist(cfg, "ask-quit-array", 2, -1, -2);
+	cfg_addlist(cfg, "ask-quit-array", 2, 1, 2);
 
 	for(i = 0; i < cfg_size(cfg, "ask-quit-array"); i++)
 		printf("ask-quit-array[%d] == %ld\n",
 			   i, cfg_getnint(cfg, "ask-quit-array", i));
 
+	/* print the parsed values to another file */
+	{
+		FILE *fp = fopen("test.conf.out", "w");
+		cfg_set_print_func(cfg, "func", print_func);
+		cfg_set_print_func(cfg, "ask-quit", print_ask);
+		cfg_set_print_func(cfg, "ask-quit-array", print_ask);
+		cfg_print(cfg, fp);
+		fclose(fp);
+	}
+
 	cfg_free(cfg);
 	return 0;
 }
