@@ -4,7 +4,7 @@
 int main(void)
 {
 	cfg_bool_t verbose = cfg_false;
-	char *server = strdup("gazonk");
+	char *server = NULL;
 	double delay = 1.356e-32;
 	char *username = NULL;
 	int debug = 1;
@@ -19,6 +19,9 @@ int main(void)
 	};
 	cfg_t *cfg;
 
+	/* set default value for the server option */
+	server = strdup("gazonk");
+
 	cfg = cfg_init(opts, 0);
 	cfg_parse(cfg, "simple.conf");
 
@@ -37,6 +40,13 @@ int main(void)
 	cfg_setstr(cfg, "user", "foo");
 	printf("username: %s\n", username);
 
+	/* print the parsed values to another file */
+	{
+		FILE *fp = fopen("simple.conf.out", "w");
+		cfg_print(cfg, fp);
+		fclose(fp);
+	}
+
 	cfg_free(cfg);
 	return 0;
 }
