@@ -1,4 +1,6 @@
 /* Configuration file parser -*- tab-width: 4; -*-
+ *
+ * $Id: confuse.c,v 1.18 2003/09/23 14:59:21 mhe Exp $
  *
  * Copyright (c) 2002-2003, Martin Hedenfalk <mhe@home.se>
  *
@@ -41,9 +43,10 @@
 
 #include "confuse.h"
 
-#if ENABLE_NLS
+#define is_set(f, x) (((f) & (x)) == (f))
+
+#if defined(ENABLE_NLS) && defined(HAVE_GETTEXT)
 # include <libintl.h>
-# include <locale.h>
 # define _(str) dgettext(PACKAGE, str)
 #else
 # define _(str) str
@@ -67,6 +70,7 @@ const char confuse_author[] = "Martin Hedenfalk <mhe@home.se>";
 
 static int cfg_parse_internal(cfg_t *cfg, int level,
 							  int force_state, cfg_opt_t *force_opt);
+static cfg_value_t *cfg_setopt(cfg_t *cfg, cfg_opt_t *opt, char *value);
 
 #define STATE_CONTINUE 0
 #define STATE_EOF -1
@@ -100,12 +104,13 @@ static char *strndup(const char *s, size_t n)
 
 	r = (char *)malloc(n + 1);
 	strncpy(r, s, n);
+	r[n] = 0;
 	return r;
 }
 #endif
 
 #ifndef HAVE_STRCASECMP
-/* Implementation from from GNU glibc 2.3.1
+/* Implementation from GNU glibc 2.3.1
  * Copyright (C) 1991,1992,1995,1996,1997,2001,2002
  *   Free Software Foundation, Inc.
  * License: GNU LGPL
@@ -129,17 +134,37 @@ static int strcasecmp(const char *a, const char *b)
 DLLIMPORT cfg_opt_t *cfg_getopt(cfg_t *cfg, const char *name)
 {
 	unsigned int i;
+	cfg_t *sec = cfg;
 
 	assert(cfg && cfg->name && name);
 
-	for(i = 0; cfg->opts[i].name; i++) {
-		if(is_set(CFGF_NOCASE, cfg->flags)) {
-			if(strcasecmp(cfg->opts[i].name, name) == 0) {
-				return &cfg->opts[i];
+	while(name && *name) {
+		char *secname;
+		size_t len = strcspn(name, "|");
+		if(name[len] == 0 /*len == strlen(name)*/)
+			/* no more subsections */
+			break;
+		if(len) {
+			secname = strndup(name, len);
+			sec = cfg_getsec(sec, secname);
+			if(sec == 0)
+				cfg_error(cfg, _("no such option '%s'"), secname);
+			free(secname);
+			if(sec == 0)
+				return 0;
+		}
+		name += len;
+		name += strspn(name, "|");
+	}
+
+	for(i = 0; sec->opts[i].name; i++) {
+		if(is_set(CFGF_NOCASE, sec->flags)) {
+			if(strcasecmp(sec->opts[i].name, name) == 0) {
+				return &sec->opts[i];
 			}
 		} else {
-			if(strcmp(cfg->opts[i].name, name) == 0) {
-				return &cfg->opts[i];
+			if(strcmp(sec->opts[i].name, name) == 0) {
+				return &sec->opts[i];
 			}
 		}
 	}
@@ -152,24 +177,33 @@ DLLIMPORT const char *cfg_title(cfg_t *cfg)
 	return cfg->title;
 }
 
-DLLIMPORT unsigned int cfg_size(cfg_t *cfg, const char *name)
+DLLIMPORT unsigned int cfg_opt_size(cfg_opt_t *opt)
 {
-	cfg_opt_t *opt = cfg_getopt(cfg, name);
 	if(opt)
 		return opt->nvalues;
 	return 0;
 }
 
+DLLIMPORT unsigned int cfg_size(cfg_t *cfg, const char *name)
+{
+	return cfg_opt_size(cfg_getopt(cfg, name));
+}
+
+DLLIMPORT signed long cfg_opt_getnint(cfg_opt_t *opt, unsigned int index)
+{
+	assert(opt && opt->type == CFGT_INT);
+	if(opt->values && index < opt->nvalues)
+		return opt->values[index]->number;
+	else if(opt->simple_value)
+		return *(signed long *)opt->simple_value;
+	else
+		return 0;
+}
+
 DLLIMPORT signed long cfg_getnint(cfg_t *cfg, const char *name,
 								  unsigned int index)
 {
-	cfg_opt_t *opt = cfg_getopt(cfg, name);
-	if(opt) {
-		assert(opt->type == CFGT_INT);
-		assert(index < opt->nvalues);
-		return opt->values[index]->number;
-	} else
-		return 0;
+	return cfg_opt_getnint(cfg_getopt(cfg, name), index);
 }
 
 DLLIMPORT signed long cfg_getint(cfg_t *cfg, const char *name)
@@ -177,16 +211,21 @@ DLLIMPORT signed long cfg_getint(cfg_t *cfg, const char *name)
 	return cfg_getnint(cfg, name, 0);
 }
 
+DLLIMPORT double cfg_opt_getnfloat(cfg_opt_t *opt, unsigned int index)
+{
+	assert(opt && opt->type == CFGT_FLOAT);
+	if(opt->values && index < opt->nvalues)
+		return opt->values[index]->fpnumber;
+	else if(opt->simple_value)
+		return *(double *)opt->simple_value;
+	else
+		return 0;
+}
+
 DLLIMPORT double cfg_getnfloat(cfg_t *cfg, const char *name,
 							   unsigned int index)
 {
-	cfg_opt_t *opt = cfg_getopt(cfg, name);
-	if(opt) {
-		assert(opt->type == CFGT_FLOAT);
-		assert(index < opt->nvalues);
-		return opt->values[index]->fpnumber;
-	} else
-		return 0;
+	return cfg_opt_getnfloat(cfg_getopt(cfg, name), index);
 }
 
 DLLIMPORT double cfg_getfloat(cfg_t *cfg, const char *name)
@@ -194,16 +233,21 @@ DLLIMPORT double cfg_getfloat(cfg_t *cfg, const char *name)
 	return cfg_getnfloat(cfg, name, 0);
 }
 
+DLLIMPORT cfg_bool_t cfg_opt_getnbool(cfg_opt_t *opt, unsigned int index)
+{
+	assert(opt && opt->type == CFGT_BOOL);
+	if(opt->values && index < opt->nvalues)
+		return opt->values[index]->boolean;
+	else if(opt->simple_value)
+		return *(cfg_bool_t *)opt->simple_value;
+	else
+		return cfg_false;
+}
+
 DLLIMPORT cfg_bool_t cfg_getnbool(cfg_t *cfg, const char *name,
 								  unsigned int index)
 {
-	cfg_opt_t *opt = cfg_getopt(cfg, name);
-	if(opt) {
-		assert(opt->type == CFGT_BOOL);
-		assert(index < opt->nvalues);
-		return opt->values[index]->boolean;
-	} else
-		return cfg_false;
+	return cfg_opt_getnbool(cfg_getopt(cfg, name), index);
 }
 
 DLLIMPORT cfg_bool_t cfg_getbool(cfg_t *cfg, const char *name)
@@ -211,15 +255,20 @@ DLLIMPORT cfg_bool_t cfg_getbool(cfg_t *cfg, const char *name)
 	return cfg_getnbool(cfg, name, 0);
 }
 
+DLLIMPORT char *cfg_opt_getnstr(cfg_opt_t *opt, unsigned int index)
+{
+	assert(opt && opt->type == CFGT_STR);
+	if(opt->values && index < opt->nvalues)
+		return opt->values[index]->string;
+	else if(opt->simple_value)
+		return *(char **)opt->simple_value;
+	else
+		return 0;
+}
+
 DLLIMPORT char *cfg_getnstr(cfg_t *cfg, const char *name, unsigned int index)
 {
-	cfg_opt_t *opt = cfg_getopt(cfg, name);
-	if(opt) {
-		assert(opt->type == CFGT_STR);
-		assert(index < opt->nvalues);
-		return opt->values[index]->string;
-	}
-	return 0;
+	return cfg_opt_getnstr(cfg_getopt(cfg, name), index);
 }
 
 DLLIMPORT char *cfg_getstr(cfg_t *cfg, const char *name)
@@ -227,14 +276,34 @@ DLLIMPORT char *cfg_getstr(cfg_t *cfg, const char *name)
 	return cfg_getnstr(cfg, name, 0);
 }
 
+DLLIMPORT cfg_t *cfg_opt_getnsec(cfg_opt_t *opt, unsigned int index)
+{
+	assert(opt && opt->type == CFGT_SEC);
+	if(opt->values && index < opt->nvalues)
+		return opt->values[index]->section;
+	return 0;
+}
+
 DLLIMPORT cfg_t *cfg_getnsec(cfg_t *cfg, const char *name, unsigned int index)
 {
-	cfg_opt_t *opt = cfg_getopt(cfg, name);
-	if(opt) {
-		assert(opt->type == CFGT_SEC);
-		assert(opt->values);
-		assert(index < opt->nvalues);
-		return opt->values[index]->section;
+	return cfg_opt_getnsec(cfg_getopt(cfg, name), index);
+}
+
+DLLIMPORT cfg_t *cfg_opt_gettsec(cfg_opt_t *opt, const char *title)
+{
+	unsigned int i, n;
+
+	n = cfg_opt_size(opt);
+	for(i = 0; i < n; i++) {
+		cfg_t *sec = cfg_opt_getnsec(opt, i);
+		assert(sec && sec->title);
+		if(is_set(CFGF_NOCASE, opt->flags)) {
+			if(strcasecmp(title, sec->title) == 0)
+				return sec;
+		} else {
+			if(strcmp(title, sec->title) == 0)
+				return sec;
+		}
 	}
 	return 0;
 }
@@ -303,11 +372,26 @@ static void cfg_init_defaults(cfg_t *cfg)
 	int i;
 
 	for(i = 0; cfg->opts[i].name; i++) {
+		/* libConfuse doesn't handle default values for "simple" options */
+		if(cfg->opts[i].simple_value)
+			continue;
+
 		if(cfg->opts[i].type != CFGT_SEC) {
 
-			if(cfg->opts[i].def.parsed) {
+			cfg->opts[i].flags |= CFGF_DEFINIT;
+
+			if(is_set(CFGF_LIST, cfg->opts[i].flags) ||
+			   cfg->opts[i].def.parsed)
+			{
 				int xstate, ret;
 
+				/* If it's a list, but no default value was given,
+				 * keep the option uninitialized.
+				 */
+				if(cfg->opts[i].def.parsed == 0 ||
+				   cfg->opts[i].def.parsed[0] == 0)
+					continue;
+
 				/* setup scanning from the string specified for the
 				 * "default" value, force the correct state and option
 				 */
@@ -344,19 +428,19 @@ static void cfg_init_defaults(cfg_t *cfg)
 			} else {
 				switch(cfg->opts[i].type) {
 					case CFGT_INT:
-						cfg_opt_setnint(cfg, &cfg->opts[i],
+						cfg_opt_setnint(&cfg->opts[i],
 										cfg->opts[i].def.number, 0);
 						break;
 					case CFGT_FLOAT:
-						cfg_opt_setnfloat(cfg, &cfg->opts[i],
+						cfg_opt_setnfloat(&cfg->opts[i],
 										  cfg->opts[i].def.fpnumber, 0);
 						break;
 					case CFGT_BOOL:
-						cfg_opt_setnbool(cfg, &cfg->opts[i],
+						cfg_opt_setnbool(&cfg->opts[i],
 										 cfg->opts[i].def.boolean, 0);
 						break;
 					case CFGT_STR:
-						cfg_opt_setnstr(cfg, &cfg->opts[i],
+						cfg_opt_setnstr(&cfg->opts[i],
 										cfg->opts[i].def.string, 0);
 						break;
 					case CFGT_FUNC:
@@ -375,6 +459,9 @@ static void cfg_init_defaults(cfg_t *cfg)
 			 * will be freed and the flag unset.
 			 */
 			cfg->opts[i].flags |= CFGF_RESET;
+		} else if(!is_set(CFGF_MULTI, cfg->opts[i].flags)) {
+			cfg_setopt(cfg, &cfg->opts[i], 0);
+			cfg->opts[i].flags |= CFGF_DEFINIT;
 		}
 	}
 }
@@ -427,8 +514,8 @@ static cfg_value_t *cfg_setopt(cfg_t *cfg, cfg_opt_t *opt, char *value)
 
 	switch(opt->type) {
 		case CFGT_INT:
-			if(opt->cb) {
-				if((*opt->cb)(cfg, opt, value, &i) != 0)
+			if(opt->parsecb) {
+				if((*opt->parsecb)(cfg, opt, value, &i) != 0)
 					return 0;
 				val->number = i;
 			} else {
@@ -447,8 +534,8 @@ static cfg_value_t *cfg_setopt(cfg_t *cfg, cfg_opt_t *opt, char *value)
 			}
 			break;
 		case CFGT_FLOAT:
-			if(opt->cb) {
-				if((*opt->cb)(cfg, opt, value, &f) != 0)
+			if(opt->parsecb) {
+				if((*opt->parsecb)(cfg, opt, value, &f) != 0)
 					return 0;
 				val->fpnumber = f;
 			} else {
@@ -469,31 +556,36 @@ static cfg_value_t *cfg_setopt(cfg_t *cfg, cfg_opt_t *opt, char *value)
 			break;
 		case CFGT_STR:
 			free(val->string);
-			if(opt->cb) {
+			if(opt->parsecb) {
 				s = 0;
-				if((*opt->cb)(cfg, opt, value, &s) != 0)
+				if((*opt->parsecb)(cfg, opt, value, &s) != 0)
 					return 0;
 				val->string = strdup(s);
 			} else
 				val->string = strdup(value);
 			break;
 		case CFGT_SEC:
-			val->section = (cfg_t *)malloc(sizeof(cfg_t));
-			assert(val->section);
-			memset(val->section, 0, sizeof(cfg_t));
-			val->section->name = strdup(opt->name);
-			val->section->opts = cfg_dupopts(opt->subopts);
-			val->section->flags = cfg->flags;
-			val->section->flags |= CFGF_ALLOCATED;
-			val->section->filename = cfg->filename;
-			val->section->line = cfg->line;
-			val->section->errfunc = cfg->errfunc;
-			val->section->title = value;
-			cfg_init_defaults(val->section);
+			if(is_set(CFGF_MULTI, opt->flags) || val->section == 0)
+			{
+				cfg_free(val->section);
+				val->section = (cfg_t *)malloc(sizeof(cfg_t));
+				assert(val->section);
+				memset(val->section, 0, sizeof(cfg_t));
+				val->section->name = strdup(opt->name);
+				val->section->opts = cfg_dupopts(opt->subopts);
+				val->section->flags = cfg->flags;
+				val->section->flags |= CFGF_ALLOCATED;
+				val->section->filename = cfg->filename;
+				val->section->line = cfg->line;
+				val->section->errfunc = cfg->errfunc;
+				val->section->title = value;
+			}
+			if(!is_set(CFGF_DEFINIT, opt->flags))
+				cfg_init_defaults(val->section);
 			break;
 		case CFGT_BOOL:
-			if(opt->cb) {
-				if((*opt->cb)(cfg, opt, value, &b) != 0)
+			if(opt->parsecb) {
+				if((*opt->parsecb)(cfg, opt, value, &b) != 0)
 					return 0;
 			} else {
 				b = cfg_parse_boolean(value);
@@ -521,14 +613,16 @@ DLLIMPORT void cfg_free_value(cfg_opt_t *opt)
 	if(opt == 0)
 		return;
 
-	for(i = 0; i < opt->nvalues; i++) {
-		if(opt->type == CFGT_STR)
-			free(opt->values[i]->string);
-		else if(opt->type == CFGT_SEC)
-			cfg_free(opt->values[i]->section);
-		free(opt->values[i]);
+	if(opt->values) {
+		for(i = 0; i < opt->nvalues; i++) {
+			if(opt->type == CFGT_STR)
+				free(opt->values[i]->string);
+			else if(opt->type == CFGT_SEC)
+				cfg_free(opt->values[i]->section);
+			free(opt->values[i]);
+		}
+		free(opt->values);
 	}
-	free(opt->values);
 	opt->values = 0;
 	opt->nvalues = 0;
 }
@@ -591,6 +685,8 @@ static int cfg_parse_internal(cfg_t *cfg, int level,
 	cfg_opt_t *opt = 0;
 	cfg_value_t *val = 0;
 	cfg_opt_t funcopt = CFG_STR(0, 0, 0);
+	int num_values = 0; /* number of values found for a list option */
+	int rc;
 
 	if(force_state != -1)
 		state = force_state;
@@ -647,6 +743,11 @@ static int cfg_parse_internal(cfg_t *cfg, int level,
 								  opt->name);
 						return STATE_ERROR;
 					}
+					/* Even if the reset flag was set by
+					 * cfg_init_defaults, appending to the defaults
+					 * should be ok.
+					 */
+					opt->flags &= ~CFGF_RESET;
 				} else if(tok == '=') {
 					/* set the (temporary) reset flag to clear the old
 					 * values, since we obviously didn't want to append */
@@ -656,14 +757,19 @@ static int cfg_parse_internal(cfg_t *cfg, int level,
 							  opt->name);
 					return STATE_ERROR;
 				}
-				if(is_set(CFGF_LIST, opt->flags))
+				if(is_set(CFGF_LIST, opt->flags)) {
 					state = 3;
-				else
+					num_values = 0;
+				} else
 					state = 2;
 				break;
 			case 2: /* expecting an option value */
 				if(tok == '}' && is_set(CFGF_LIST, opt->flags)) {
 					state = 0;
+					if(num_values == 0 && is_set(CFGF_RESET, opt->flags))
+						/* Reset flags was set, and the empty list was
+						 * specified. Free all old values. */
+						cfg_free_value(opt);
 					break;
 				}
 
@@ -674,9 +780,12 @@ static int cfg_parse_internal(cfg_t *cfg, int level,
 
 				if(cfg_setopt(cfg, opt, cfg_yylval) == 0)
 					return STATE_ERROR;
-				if(is_set(CFGF_LIST, opt->flags))
+				if(opt->validcb && (*opt->validcb)(cfg, opt) != 0)
+					return STATE_ERROR;
+				if(is_set(CFGF_LIST, opt->flags)) {
 					state = 4;
-				else
+					++num_values;
+				} else
 					state = 0;
 				break;
 			case 3: /* expecting an opening brace for a list option */
@@ -691,9 +800,11 @@ static int cfg_parse_internal(cfg_t *cfg, int level,
 					 * closing (list) brace */
 				if(tok == ',')
 					state = 2;
-				else if(tok == '}')
+				else if(tok == '}') {
 					state = 0;
-				else {
+					if(opt->validcb && (*opt->validcb)(cfg, opt) != 0)
+						return STATE_ERROR;
+				} else {
 					cfg_error(cfg, _("unexpected token '%s'"), cfg_yylval);
 					return STATE_ERROR;
 				}
@@ -710,9 +821,13 @@ static int cfg_parse_internal(cfg_t *cfg, int level,
 				if(!val)
 					return STATE_ERROR;
 
-				if(cfg_parse_internal(val->section, level+1,-1,0) != STATE_EOF)
-					return STATE_ERROR;
+				val->section->line = cfg->line;
+				rc = cfg_parse_internal(val->section, level+1,-1,0);
 				cfg->line = val->section->line;
+				if(rc != STATE_EOF)
+					return STATE_ERROR;
+				if(opt->validcb && (*opt->validcb)(cfg, opt) != 0)
+					return STATE_ERROR;
 				state = 0;
 				break;
 			case 6: /* expecting a title for a section */
@@ -796,12 +911,13 @@ DLLIMPORT int cfg_parse(cfg_t *cfg, const char *filename)
 
 	assert(cfg && filename);
 
+	free(cfg->filename);
 	cfg->filename = cfg_tilde_expand(filename);
 	fp = fopen(cfg->filename, "r");
 	if(fp == 0)
 		return CFG_FILE_ERROR;
 	ret = cfg_parse_fp(cfg, fp);
-	fclose(cfg_yyin);
+	fclose(fp);
 	return ret;
 }
 
@@ -809,8 +925,8 @@ DLLIMPORT int cfg_parse_buf(cfg_t *cfg, const char *buf)
 {
 	int ret;
 
-	if(cfg->filename == 0)
-		cfg->filename = strdup("BUF");
+	free(cfg->filename);
+	cfg->filename = strdup("[buf]");
 	cfg->line = 1;
 
 	cfg_scan_string_begin(buf);
@@ -838,7 +954,7 @@ DLLIMPORT cfg_t *cfg_init(cfg_opt_t *opts, cfg_flag_t flags)
 
 	cfg_init_defaults(cfg);
 
-#if ENABLE_NLS
+#if defined(ENABLE_NLS) && defined(HAVE_GETTEXT)
 	setlocale(LC_MESSAGES, "");
 	setlocale(LC_CTYPE, "");
 	bindtextdomain(PACKAGE, LOCALEDIR);
@@ -938,11 +1054,11 @@ static cfg_value_t *cfg_getval(cfg_opt_t *opt, unsigned int index)
 	return val;
 }
 
-DLLIMPORT void cfg_opt_setnint(cfg_t *cfg, cfg_opt_t *opt, long int value,
-					 unsigned int index)
+DLLIMPORT void cfg_opt_setnint(cfg_opt_t *opt, long int value,
+							   unsigned int index)
 {
 	cfg_value_t *val;
-	assert(cfg && opt && opt->type == CFGT_INT);
+	assert(opt && opt->type == CFGT_INT);
 	val = cfg_getval(opt, index);
 	val->number = value;
 }
@@ -950,9 +1066,7 @@ DLLIMPORT void cfg_opt_setnint(cfg_t *cfg, cfg_opt_t *opt, long int value,
 DLLIMPORT void cfg_setnint(cfg_t *cfg, const char *name,
 					 long int value, unsigned int index)
 {
-	cfg_opt_t *opt = cfg_getopt(cfg, name);
-	if(opt)
-		cfg_opt_setnint(cfg, opt, value, index);
+	cfg_opt_setnint(cfg_getopt(cfg, name), value, index);
 }
 
 DLLIMPORT void cfg_setint(cfg_t *cfg, const char *name, long int value)
@@ -960,11 +1074,11 @@ DLLIMPORT void cfg_setint(cfg_t *cfg, const char *name, long int value)
 	cfg_setnint(cfg, name, value, 0);
 }
 
-DLLIMPORT void cfg_opt_setnfloat(cfg_t *cfg, cfg_opt_t *opt, double value,
-					   unsigned int index)
+DLLIMPORT void cfg_opt_setnfloat(cfg_opt_t *opt, double value,
+								 unsigned int index)
 {
 	cfg_value_t *val;
-	assert(cfg && opt && opt->type == CFGT_FLOAT);
+	assert(opt && opt->type == CFGT_FLOAT);
 	val = cfg_getval(opt, index);
 	val->fpnumber = value;
 }
@@ -972,9 +1086,7 @@ DLLIMPORT void cfg_opt_setnfloat(cfg_t *cfg, cfg_opt_t *opt, double value,
 DLLIMPORT void cfg_setnfloat(cfg_t *cfg, const char *name,
 				   double value, unsigned int index)
 {
-	cfg_opt_t *opt = cfg_getopt(cfg, name);
-	if(opt)
-		cfg_opt_setnfloat(cfg, opt, value, index);
+	cfg_opt_setnfloat(cfg_getopt(cfg, name), value, index);
 }
 
 DLLIMPORT void cfg_setfloat(cfg_t *cfg, const char *name, double value)
@@ -982,21 +1094,19 @@ DLLIMPORT void cfg_setfloat(cfg_t *cfg, const char *name, double value)
 	cfg_setnfloat(cfg, name, value, 0);
 }
 
-DLLIMPORT void cfg_opt_setnbool(cfg_t *cfg, cfg_opt_t *opt, cfg_bool_t value,
-					  unsigned int index)
+DLLIMPORT void cfg_opt_setnbool(cfg_opt_t *opt, cfg_bool_t value,
+								unsigned int index)
 {
 	cfg_value_t *val;
-	assert(cfg && opt && opt->type == CFGT_BOOL);
+	assert(opt && opt->type == CFGT_BOOL);
 	val = cfg_getval(opt, index);
 	val->boolean = value;
 }
 
 DLLIMPORT void cfg_setnbool(cfg_t *cfg, const char *name,
-				  cfg_bool_t value, unsigned int index)
+							cfg_bool_t value, unsigned int index)
 {
-	cfg_opt_t *opt = cfg_getopt(cfg, name);
-	if(opt)
-		cfg_opt_setnbool(cfg, opt, value, index);
+	cfg_opt_setnbool(cfg_getopt(cfg, name), value, index);
 }
 
 DLLIMPORT void cfg_setbool(cfg_t *cfg, const char *name, cfg_bool_t value)
@@ -1004,22 +1114,20 @@ DLLIMPORT void cfg_setbool(cfg_t *cfg, const char *name, cfg_bool_t value)
 	cfg_setnbool(cfg, name, value, 0);
 }
 
-DLLIMPORT void cfg_opt_setnstr(cfg_t *cfg, cfg_opt_t *opt, const char *value,
-					 unsigned int index)
+DLLIMPORT void cfg_opt_setnstr(cfg_opt_t *opt, const char *value,
+							   unsigned int index)
 {
 	cfg_value_t *val;
-	assert(cfg && opt && opt->type == CFGT_STR);
+	assert(opt && opt->type == CFGT_STR);
 	val = cfg_getval(opt, index);
 	free(val->string);
 	val->string = value ? strdup(value) : 0;
 }
 
 DLLIMPORT void cfg_setnstr(cfg_t *cfg, const char *name,
-				 const char *value, unsigned int index)
+						   const char *value, unsigned int index)
 {
-	cfg_opt_t *opt = cfg_getopt(cfg, name);
-	if(opt)
-		cfg_opt_setnstr(cfg, opt, value, index);
+	cfg_opt_setnstr(cfg_getopt(cfg, name), value, index);
 }
 
 DLLIMPORT void cfg_setstr(cfg_t *cfg, const char *name, const char *value)
@@ -1027,7 +1135,7 @@ DLLIMPORT void cfg_setstr(cfg_t *cfg, const char *name, const char *value)
 	cfg_setnstr(cfg, name, value, 0);
 }
 
-static void cfg_addlist_internal(cfg_t *cfg, cfg_opt_t *opt,
+static void cfg_addlist_internal(cfg_opt_t *opt,
 								 unsigned int nvalues, va_list ap)
 {
 	unsigned int i;
@@ -1035,18 +1143,18 @@ static void cfg_addlist_internal(cfg_t *cfg, cfg_opt_t *opt,
 	for(i = 0; i < nvalues; i++) {
 		switch(opt->type) {
 			case CFGT_INT:
-				cfg_opt_setnint(cfg, opt, va_arg(ap, int), opt->nvalues);
+				cfg_opt_setnint(opt, va_arg(ap, int), opt->nvalues);
 				break;
 			case CFGT_FLOAT:
-				cfg_opt_setnfloat(cfg, opt, va_arg(ap, double),
+				cfg_opt_setnfloat(opt, va_arg(ap, double),
 								  opt->nvalues);
 				break;
 			case CFGT_BOOL:
-				cfg_opt_setnbool(cfg, opt, va_arg(ap, cfg_bool_t),
+				cfg_opt_setnbool(opt, va_arg(ap, cfg_bool_t),
 								 opt->nvalues);
 				break;
 			case CFGT_STR:
-				cfg_opt_setnstr(cfg, opt, va_arg(ap, char*), opt->nvalues);
+				cfg_opt_setnstr(opt, va_arg(ap, char*), opt->nvalues);
 				break;
 			case CFGT_FUNC:
 			case CFGT_SEC:
@@ -1059,29 +1167,173 @@ static void cfg_addlist_internal(cfg_t *cfg, cfg_opt_t *opt,
 DLLIMPORT void cfg_setlist(cfg_t *cfg, const char *name,
 						   unsigned int nvalues, ...)
 {
+	va_list ap;
 	cfg_opt_t *opt = cfg_getopt(cfg, name);
 
-	if(opt) {
-		va_list ap;
+	assert(opt && is_set(CFGF_LIST, opt->flags));
 
-		cfg_free_value(opt);
-		va_start(ap, nvalues);
-		cfg_addlist_internal(cfg, opt, nvalues, ap);
-		va_end(ap);
-	}
+	cfg_free_value(opt);
+	va_start(ap, nvalues);
+	cfg_addlist_internal(opt, nvalues, ap);
+	va_end(ap);
 }
 
 DLLIMPORT void cfg_addlist(cfg_t *cfg, const char *name,
 						   unsigned int nvalues, ...)
 {
+	va_list ap;
 	cfg_opt_t *opt = cfg_getopt(cfg, name);
 
-	if(opt) {
-		va_list ap;
+	assert(opt && is_set(CFGF_LIST, opt->flags));
 
-		va_start(ap, nvalues);
-		cfg_addlist_internal(cfg, opt, nvalues, ap);
-		va_end(ap);
+	va_start(ap, nvalues);
+	cfg_addlist_internal(opt, nvalues, ap);
+	va_end(ap);
+}
+
+DLLIMPORT void cfg_opt_nprint_var(cfg_opt_t *opt, unsigned int index, FILE *fp)
+{
+	const char *str;
+
+	switch(opt->type)
+	{
+		case CFGT_INT:
+			fprintf(fp, "%ld", cfg_opt_getnint(opt, index));
+			break;
+		case CFGT_FLOAT:
+			fprintf(fp, "%lf", cfg_opt_getnfloat(opt, index));
+			break;
+		case CFGT_STR:
+			str = cfg_opt_getnstr(opt, index);
+			fprintf(fp, "\"%s\"", str ? str : "");
+			break;
+		case CFGT_BOOL:
+			fprintf(fp, "%s", cfg_opt_getnbool(opt, index) ? "true" : "false");
+			break;
+		case CFGT_NONE:
+		case CFGT_SEC:
+		case CFGT_FUNC:
+			break;
 	}
 }
 
+static void cfg_indent(FILE *fp, int indent)
+{
+	while(indent--)
+		fprintf(fp, "  ");
+}
+
+DLLIMPORT void cfg_opt_print_indent(cfg_opt_t *opt, FILE *fp, int indent)
+{
+	assert(opt);
+
+	if(opt->type == CFGT_SEC) {
+		cfg_t *sec;
+		unsigned int i;
+
+		for(i = 0; i < cfg_opt_size(opt); i++) {
+			sec = cfg_opt_getnsec(opt, i);
+			cfg_indent(fp, indent);
+			if(is_set(CFGF_TITLE, opt->flags))
+				fprintf(fp, "%s \"%s\" {\n", opt->name, cfg_title(sec));
+			else
+				fprintf(fp, "%s {\n", opt->name);
+			cfg_print_indent(sec, fp, indent + 1);
+			cfg_indent(fp, indent);
+			fprintf(fp, "}\n");
+		}
+	} else if(opt->type != CFGT_FUNC && opt->type != CFGT_NONE) {
+		if(is_set(CFGF_LIST, opt->flags)) {
+			unsigned int i;
+
+			cfg_indent(fp, indent);
+			fprintf(fp, "%s = {", opt->name);
+
+			if(opt->nvalues) {
+				if(opt->pf)
+					opt->pf(opt, 0, fp);
+				else
+					cfg_opt_nprint_var(opt, 0, fp);
+				for(i = 1; i < opt->nvalues; i++) {
+					fprintf(fp, ", ");
+					if(opt->pf)
+						opt->pf(opt, i, fp);
+					else
+						cfg_opt_nprint_var(opt, i, fp);
+				}
+			}
+
+			fprintf(fp, "}");
+		} else {
+			cfg_indent(fp, indent);
+			/* comment out the option if is not set */
+			if(opt->simple_value) {
+				if(opt->type == CFGT_STR && *((char **)opt->simple_value) == 0)
+					fprintf(fp, "# ");
+			} else {
+				if(cfg_opt_size(opt) == 0 || (
+					   opt->type == CFGT_STR && (opt->values[0]->string == 0 ||
+											 opt->values[0]->string[0] == 0)))
+					fprintf(fp, "# ");
+			}
+			fprintf(fp, "%s = ", opt->name);
+			if(opt->pf)
+				opt->pf(opt, 0, fp);
+			else
+				cfg_opt_nprint_var(opt, 0, fp);
+		}
+	
+		fprintf(fp, "\n");
+	} else if(opt->pf) {
+		cfg_indent(fp, indent);
+		opt->pf(opt, 0, fp);
+		fprintf(fp, "\n");
+	}
+}
+
+DLLIMPORT void cfg_opt_print(cfg_opt_t *opt, FILE *fp)
+{
+	cfg_opt_print_indent(opt, fp, 0);
+}
+
+DLLIMPORT void cfg_print_indent(cfg_t *cfg, FILE *fp, int indent)
+{
+	int i;
+
+	for(i = 0; cfg->opts[i].name; i++)
+		cfg_opt_print_indent(&cfg->opts[i], fp, indent);
+}
+
+DLLIMPORT void cfg_print(cfg_t *cfg, FILE *fp)
+{
+	cfg_print_indent(cfg, fp, 0);
+}
+
+DLLIMPORT cfg_print_func_t cfg_opt_set_print_func(cfg_opt_t *opt,
+												  cfg_print_func_t pf)
+{
+	cfg_print_func_t oldpf;
+
+	assert(opt);
+	oldpf = opt->pf;
+	opt->pf = pf;
+
+	return oldpf;
+}
+
+DLLIMPORT cfg_print_func_t cfg_set_print_func(cfg_t *cfg, const char *name,
+											  cfg_print_func_t pf)
+{
+	return cfg_opt_set_print_func(cfg_getopt(cfg, name), pf);
+}
+
+DLLIMPORT cfg_validate_callback_t cfg_set_validate_func(cfg_t *cfg,
+														const char *name,
+												   cfg_validate_callback_t vf)
+{
+	cfg_opt_t *opt = cfg_getopt(cfg, name);
+	assert(opt);
+	cfg_validate_callback_t oldvf = opt->validcb;
+	opt->validcb = vf;
+	return oldvf;
+}
